from functools import wraps
from typing import Callable,Dict
from loguru import logger


PON_METHOD_ATTR_NAME = 'pon_consumer_func_config'

PON_CONSUMER_METHODS: Dict[type, Callable] = {}


def event_handler(source_service: str, event_name: str, **event_handler_kwagrs):
    def decorate(pon_service_cls_consumer_method: Callable):

        pon_consumer_func_config = {
            'source_service': source_service,
            'event_name': event_name
        }

        pon_consumer_func_config.update(event_handler_kwagrs)

        if not hasattr(pon_service_cls_consumer_method, PON_METHOD_ATTR_NAME):
            setattr(
                pon_service_cls_consumer_method,
                PON_METHOD_ATTR_NAME,
                pon_consumer_func_config
            )

        @wraps(pon_service_cls_consumer_method)
        def wrapper(*args, **kwargs):
            return pon_service_cls_consumer_method(*args, **kwargs)
        return wrapper
    return decorate



from typing import Union, Dict
from kombu import Exchange
from kombu import Connection
from kombu.messaging import Producer
from kombu.transport.base import Message

from kombu import Exchange, Queue
from loguru import logger
import time
from pon.constants import NON_PERSISTENT, PERSISTENT


def get_event_exchange(service_name: str) -> Exchange:
    exchange_name = "{}.events".format(service_name)
    exchange = Exchange(
        exchange_name,
        type='topic',
        durable=True,
        delivery_mode=PERSISTENT,
        auto_delete=False,
        no_declare=False,
    )

    return exchange


def event_dispatcher(amqp_uri: str, **default_kwargs):
    from pon.amqp.pool.producer import get_producer_from_pool_group

    def dispatch(service_name: str, event_type: str, event_data: Union[str, bytes, Dict], **kwargs):
        default_kwargs.update(kwargs)
        """ Dispatch an event claiming to originate from `service_name` with
        the given `event_type` and `event_data`.
        """
        exchange = get_event_exchange(service_name)
        routing_key: str = event_type

        with get_producer_from_pool_group(Connection(amqp_uri)) as producer:
            message = Message(body=event_data)
            producer.publish(
                body=message.body,
                exchange=exchange,
                routing_key=routing_key,
                **default_kwargs  # TODO args safe
            )
    return dispatch



import os
import re
from functools import partial

import yaml


has_regex_module = False
ENV_VAR_MATCHER = re.compile(
    r"""
        \$\{       # match characters `${` literally
        ([^}:\s]+) # 1st group: matches any character except `}` or `:`
        :?         # matches the literal `:` character zero or one times
        ([^}]+)?   # 2nd group: matches any character except `}`
        \}         # match character `}` literally
    """, re.VERBOSE
)


IMPLICIT_ENV_VAR_MATCHER = re.compile(
    r"""
        .*          # matches any number of any characters
        \$\{.*\}    # matches any number of any characters
                    # between `${` and `}` literally
        .*          # matches any number of any characters
    """, re.VERBOSE
)


RECURSIVE_ENV_VAR_MATCHER = re.compile(
    r"""
        \$\{       # match characters `${` literally
        ([^}]+)?   # matches any character except `}`
        \}         # match character `}` literally
        ([^$}]+)?  # matches any character except `}` or `$`
        \}         # match character `}` literally
    """,
    re.VERBOSE,
)


def _replace_env_var(match):
    env_var, default = match.groups()
    value = os.environ.get(env_var, None)
    if value is None:
        # expand default using other vars
        if default is None:
            # regex module return None instead of
            #  '' if engine didn't entered default capture group
            default = ''

        value = default
        while IMPLICIT_ENV_VAR_MATCHER.match(value):  # pragma: no cover
            value = ENV_VAR_MATCHER.sub(_replace_env_var, value)
    return value


def env_var_constructor(loader, node, raw=False):
    raw_value = loader.construct_scalar(node)

    # detect and error on recursive environment variables
    if not has_regex_module and RECURSIVE_ENV_VAR_MATCHER.match(
        raw_value
    ):  # pragma: no cover
        raise Exception(
            "Nested environment variable lookup requires the `regex` module"
        )
    value = ENV_VAR_MATCHER.sub(_replace_env_var, raw_value)
    if value == raw_value:
        return value  # avoid recursion
    return value if raw else yaml.safe_load(value)


def setup_yaml_parser():
    yaml.add_constructor('!env_var', env_var_constructor, yaml.SafeLoader)
    yaml.add_constructor(
        '!raw_env_var',
        partial(env_var_constructor, raw=True),
        yaml.SafeLoader
    )
    yaml.add_implicit_resolver(
        '!env_var', IMPLICIT_ENV_VAR_MATCHER, Loader=yaml.SafeLoader
    )

# copy from nameko https://github.com/nameko/nameko/blob/v2.14.1/nameko/cli/main.py



from kombu import Connection
from kombu.pools import connections
from kombu.connection import ConnectionPool
from kombu import Connection


def get_connection_pool(connection: Connection) -> ConnectionPool:
    """
    connections 是 Connections 的实例, Connections 是 PoolGroup 的子类
    """
    connection_pool: ConnectionPool = connections[connection]
    return connection_pool


def get_connection_from_pool(pool: ConnectionPool) -> Connection:
    return pool.acquire(block=True)


def get_connection_from_pool_group(connection: Connection) -> Connection:
    return get_connection_from_pool(get_connection_pool(connection))



from kombu.messaging import Producer
from kombu.pools import producers, ProducerPool
from kombu import Connection, Producer


def get_producer_pool(connection: Connection) -> ProducerPool:
    producer_pool: ProducerPool = producers[connection]
    return producer_pool


def get_producer_from_pool(pool: ProducerPool) -> Producer:
    return pool.acquire(block=True)


def get_producer_from_pool_group(connection: Connection) -> Producer:
    return get_producer_from_pool(get_producer_pool(connection))



import sys
import inspect
from typing import List


def get_class_names(module_name: str) -> List[str]:
    __import__(module_name)
    module = sys.modules[module_name]

    return [
        class_name
        for class_name, class_object in
        inspect.getmembers(module, inspect.isclass)
    ]



import hashlib
from pathlib import Path


def get_file_md5(video_filepath: Path) -> str:
    """ 返回一个长度为 32 的字符串 """
    md5_encoder = hashlib.md5()
    with open(video_filepath, 'rb') as f:
        md5_encoder.update(f.read())
        return md5_encoder.hexdigest()


def get_string_md5(text: str) -> str:
    md5_encoder = hashlib.md5()
    md5_encoder.update(text.encode('utf-8'))
    return md5_encoder.hexdigest()


def get_stream_md5(stream: bytes) -> str:
    md5_encoder = hashlib.md5()
    md5_encoder.update(stream)
    return md5_encoder.hexdigest()



import inspect
from inspect import signature
from typing import (
    Any,
    Callable,
    Dict,
    ForwardRef,
    Type,
    Union,
    cast,
    get_args
)
from loguru import logger
from kombu.transport.pyamqp import Message
from kombu import Queue
from pon.events.hints import Message as PMessage
from pon.events.hints import Property, Header


class MessageEncoder:
    def encode(self, message_body: Union[str, bytes]):
        pass


def evaluate_forwardref(type_: ForwardRef, globalns: Any, localns: Any) -> Any:
    # Even though it is the right signature for python 3.9, mypy complains with
    # `error: Too many arguments for "_evaluate" of "ForwardRef"` hence the cast...
    return cast(Any, type_)._evaluate(globalns, localns, set())


def get_typed_signature(call: Callable[..., Any]) -> inspect.Signature:
    signature = inspect.signature(call)
    globalns = getattr(call, "__globals__", {})
    typed_params = [
        inspect.Parameter(
            name=param.name,
            kind=param.kind,
            default=param.default,
            annotation=get_typed_annotation(param, globalns),
        )
        for param in signature.parameters.values()
    ]
    typed_signature = inspect.Signature(typed_params)
    return typed_signature


def get_typed_annotation(param: inspect.Parameter, globalns: Dict[str, Any]) -> Any:
    annotation = param.annotation
    if isinstance(annotation, str):
        annotation = ForwardRef(annotation)
        annotation = evaluate_forwardref(annotation, globalns, globalns)
    return annotation


class MessageConsumer:

    queue: Queue
    service_cls: Type
    consumer_method: Callable

    def __init__(self, queue: Queue, service_cls: Type, consumer_method: Callable) -> None:
        self.queue = queue
        self.service_cls = service_cls
        self.consumer_method = consumer_method

    def handle_message(self, message: Message):
        """
        body 的 type 取决于 message 的 Properties 的 content_type
        content_type:
        - application/json 对应的 type 就是 dict
        """
        # assert isinstance(message.body, bytes)
        typed_signature = get_typed_signature(
            self.consumer_method).parameters.items()

        method_args = [None for _ in range(len(typed_signature))]
        try:
            for index, (parameter_name, parameter) in enumerate(typed_signature):
                if parameter_name == 'self':
                    service = self.service_cls()
                    method_args[index] = service
                if parameter_name == 'data':
                    content_encoding = message.properties.get(
                        'content_encoding', None)
                    if (str in get_args(parameter.annotation)):
                        message_body = message.body.decode(
                            content_encoding or 'utf-8')
                    else:
                        message_body = message.body
                    method_args[index] = message_body
                if parameter.annotation == PMessage:
                    method_args[index] = message

                if isinstance(parameter.default, Property):
                    method_args[index] = message.properties.get(
                        parameter_name, None)

                if isinstance(parameter.default, Header):
                    header_value = message.headers.get(parameter_name, None)
                    if header_value and (int in get_args(parameter.annotation)):
                        header_value = int(header_value)
                    method_args[index] = header_value

            self.consumer_method(*method_args)
        except Exception as error:
            logger.exception(error)
            if not message.acknowledged:
                message.reject()
        else:
            if not message.acknowledged:
                message.ack()



import os
import sys
from pathlib import Path
from typing import List, Dict, Tuple, Callable
import yaml
import eventlet
from loguru import logger
from eventlet import wsgi
from werkzeug.wrappers import Request, Response
from werkzeug.routing import Map, Rule
from werkzeug.exceptions import HTTPException
from eventlet.greenio.base import GreenSocket
from eventlet.greenthread import GreenThread


class PonApp:
    """Implements a WSGI application for managing your favorite movies."""

    def __init__(self):
        from pon.web.entrance import url_map
        self.url_map = url_map

    def dispatch_request(self, request: Request):
        """Dispatches the request."""
        # return Response('Hello World!')
        adapter = self.url_map.bind_to_environ(request.environ)
        try:
            endpoint, values = adapter.match()
            return endpoint(request, **values)
            # return getattr(self, endpoint.__name__)(request, **values)
            # return getattr(self, f'on_{endpoint}')(request, **values)
        except HTTPException as error:
            return error

    def wsgi_app(self, environ: Dict, start_response: Callable):
        """WSGI application that processes requests and returns responses."""
        request = Request(environ)
        response = self.dispatch_request(request)
        return response(environ, start_response)

    def __call__(self, environ: Dict, start_response: Callable):
        """The WSGI server calls this method as the WSGI application."""
        return self.wsgi_app(environ, start_response)


class EventletAPIRunner:
    def __init__(self) -> None:
        self.put_patch()

    def put_patch(self) -> None:
        import eventlet
        eventlet.monkey_patch()  # noqa (code before rest of imports)

    def load_service_cls_list(self, services: Tuple[str]) -> List[type]:
        cwd: Path = Path(os.getcwd())
        sys.path.append(str(cwd))

        service_cls_list: List[type] = []

        for service in services:
            items: List[str] = service.split(':')
            if len(items) == 1:
                module_name, service_class_name = items[0], None
            elif len(items) == 2:
                module_name, service_class_name = items
            else:
                raise Exception(f'错误的 service 格式: {service}')

            __import__(module_name)

            module = sys.modules[module_name]

            service_cls = getattr(module, service_class_name)

            service_cls_list.append(service_cls)

        return service_cls_list

    def load_config(self, config_filepath: Path):
        with open(config_filepath, 'r', encoding='utf-8') as file:
            config: Dict[str, Dict] = yaml.load(
                file.read(), Loader=yaml.Loader)
        self.amqp_uri = config['AMQP_URI']

    def run(self, services: Tuple[str], config_filepath: Path):
        self.load_config(config_filepath)
        # service_cls_list: List[type] = self.load_service_cls_list(services)
        self.load_service_cls_list(services)

        def create_app():
            """Application factory function that returns an instance of MovieApp."""
            app = PonApp()
            return app

        app = create_app()

        server_socket: GreenSocket = eventlet.listen(("0.0.0.0", 8000))

        # wsgi.server(server_socket, app)

        gt: GreenThread = eventlet.spawn(wsgi.server, server_socket, app)

        # gt.wait()



import os
from pathlib import Path
from typing import Tuple, Union, Optional
import click
from loguru import logger
import eventlet
import time


def run_forever():
    while True:
        time.sleep(8888)


@click.group()
def pon_cli():
    pass


@pon_cli.command()
@click.argument('services', nargs=-1, required=True)
@click.option('--config', help='config file path')
def run(services: Tuple[str], config: Optional[str] = None):
    """ 启动 pon 服务 """
    config_filepath: Path = Path(os.getcwd())/config
    from pon.events import EventletEventRunner
    from pon.web import EventletAPIRunner

    eventlet.spawn_n(EventletEventRunner().run, services, config_filepath)
    # eventlet.spawn_n(EventletAPIRunner().run, services, config_filepath)

    gt = eventlet.spawn(run_forever)
    gt.wait()


@pon_cli.command()
@click.option('--config', help='config file path')
def shell(file_path: str = None):
    """ shell 交互环境 """
    pass


cli = click.CommandCollection(sources=[pon_cli])

if __name__ == '__main__':
    cli()



from functools import wraps
from typing import Callable, List
from werkzeug.routing import Map, Rule


url_map = Map()


def http(methods: List[str], url: str):
    def decorate(service_cls_api_method: Callable):
        url_map.add(Rule(
            url, endpoint=service_cls_api_method
        ))

        @wraps(service_cls_api_method)
        def wrapper(*args, **kwargs):
            return service_cls_api_method(*args, **kwargs)
        return wrapper
    return decorate



import os
import sys
from pathlib import Path
from typing import Tuple, ClassVar, Type, Dict, List, Callable
import yaml
from loguru import logger
from kombu import Exchange, Queue
from kombu.utils.compat import nested
from kombu import Connection, Consumer, Queue
from kombu.transport.pyamqp import Channel
from pon.events.message import MessageConsumer
from pon.standalone.events import get_event_exchange
from pon.core import get_class_names


class QueueLine:
    queue: Queue
    service_cls: Type
    method: Callable

    def __init__(self, queue: Queue, service_cls: Type, method: Callable) -> None:
        self.queue = queue
        self.service_cls = service_cls
        self.method = method


class EventletEventRunner:
    amqp_uri: str
    queues: List[QueueLine]

    def __init__(self) -> None:
        self.put_patch()

    def put_patch(self) -> None:
        import eventlet
        eventlet.monkey_patch()  # noqa (code before rest of imports)

    def load_service_cls_list(self, services: Tuple[str]) -> List[type]:
        BASE_DIR: Path = Path(os.getcwd())
        sys.path.append(str(BASE_DIR))

        service_cls_list: List[type] = []

        for service in services:
            items: List[str] = service.split(':')
            if len(items) == 1:
                module_name, service_class_name = items[0], None
            elif len(items) == 2:
                module_name, service_class_name = items
            else:
                raise Exception(f'错误的 service 格式: {service}')

            __import__(module_name)
            module = sys.modules[module_name]

            if service_class_name:
                service_class_names = [service_class_name]
            else:
                service_class_names = get_class_names(module_name)

            for service_class_name in service_class_names:
                service_cls = getattr(module, service_class_name)
                service_cls_list.append(service_cls)

        return service_cls_list

    def load_config(self, config_filepath: Path):
        with open(config_filepath, 'r', encoding='utf-8') as f:
            config: Dict[str, Dict] = yaml.safe_load(f)
        self.amqp_uri = config['AMQP_URI']

    def declare_exchange(self, exchange: Exchange):
        with Connection(self.amqp_uri) as conn:
            with conn.channel() as channel:
                exchange.declare(channel=channel)

    def declare_queue(self, queue: Queue):
        with Connection(self.amqp_uri) as conn:
            with conn.channel() as channel:
                queue.declare(channel=channel)

    def run(self, services: Tuple[str], config_filepath: Path):
        self.load_config(config_filepath)

        self.queues: List[QueueLine] = []

        service_cls_list: List[type] = self.load_service_cls_list(services)

        from pon.events.entrance import PON_METHOD_ATTR_NAME
        # 1. 去 rabbitmq 创建消息队列

        for service_cls in service_cls_list:
            for item in dir(service_cls):
                cls_property: Callable = getattr(service_cls, item)
                if hasattr(cls_property, PON_METHOD_ATTR_NAME):
                    consumer_method = cls_property

                    pon_consumer_func_config = getattr(
                        consumer_method, PON_METHOD_ATTR_NAME)
                    # 获取修饰器附加的参数
                    source_service: str = pon_consumer_func_config['source_service']
                    event_name: str = pon_consumer_func_config['event_name']

                    exchange_name = get_event_exchange(
                        source_service).name
                    routing_key: str = event_name

                    exchange = Exchange(exchange_name, type='topic')

                    self.declare_exchange(exchange)

                    queue_name = f'evt-{exchange_name}-{routing_key}--{service_cls.name}.{consumer_method.__name__}'

                    queue = Queue(queue_name, exchange,
                                  routing_key, durable=True, queue_arguments={'x-max-priority': 10})
                    self.declare_queue(queue)
                    self.queues.append(QueueLine(
                        queue, service_cls, consumer_method))

        logger.info(
            f'load services: {", ".join([service_class.__name__ for service_class in service_cls_list])}')

        # 2. 开始监听和消费
        while True:
            try:
                with Connection(self.amqp_uri) as conn:
                    consumers: List[Consumer] = []
                    for queueline in self.queues:

                        channel: Channel = conn.channel()
                        consumer = Consumer(
                            channel,
                            queues=[queueline.queue],
                            prefetch_count=1,
                            on_message=MessageConsumer(
                                queue=queueline.queue,
                                service_cls=queueline.service_cls,
                                consumer_method=queueline.method
                            ).handle_message
                        )
                        consumers.append(consumer)
                    logger.info(f'start consuming {self.amqp_uri}')

                    with nested(*consumers):
                        while True:
                            conn.drain_events()
            except Exception as error:
                logger.warning(error)



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

from pon import VERSION

setuptools.setup(
    name="ponponon",
    version=VERSION,
    author="ponponon",
    author_email="1729303158@qq.com",
    maintainer='ponponon',
    maintainer_email='1729303158@qq.com',
    license='MIT License',
    platforms=["all"],
    description="An advanced message queue framework, derived from nameko",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ponponon/pon",
    packages=setuptools.find_packages(),
    install_requires=[
        "pydantic",
        "loguru",
        "pyyaml",
        "eventlet",
        "kombu",
        "werkzeug",
    ],
    extras_require={
    },
    entry_points={
        'console_scripts': [
            'pon=pon.cli.main:cli',
        ]
    },
    classifiers=[
        "Programming Language :: Python",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Developers",
    ]
)



from datetime import datetime, timedelta, timezone
import time
import contextlib


def get_min_utc_timestamp() -> datetime:
    return (datetime(year=1970, month=1, day=1) + timedelta(seconds=1)).replace(tzinfo=timezone.utc)


def get_utc_now_timestamp() -> datetime:
    """ https://blog.csdn.net/ball4022/article/details/101670024 """
    return datetime.utcnow().replace(tzinfo=timezone.utc)


def timedelta_seconds(start_time: datetime, end_time: datetime = None) -> int:
    """ 返回两个时间相差的秒数 """
    if not end_time:
        end_time = get_utc_now_timestamp()

    return int((end_time - start_time).total_seconds())


def custom_timestamp(base_timestamp: datetime, seconds: int, reduce=False):
    return base_timestamp + timedelta(seconds=seconds) \
        if not reduce \
        else base_timestamp - timedelta(seconds=seconds)


@contextlib.contextmanager
def timer(msg: str = None, logger=None):
    if not logger:
        from loguru import logger as loguru_logger
        logger = loguru_logger

    start = time.time()
    yield
    logger.debug(f'{msg}, used {round(time.time() - start,3)} s')
